// ################################################################################
//
//  HexBot2 v0.0
//
//  Released:  02/08/2021
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    microcontroller:  ESP32 Dev Kit

    This code controls a hexapod robot, giving it the ability to walk and respond to
    commands from a Wii controller over ESPNOW WiFi. The robot has a rotating head
    which contains a laser ranging device. A 64x128 OLED display provides
    feedback on modes and battery performance. RGB LEDs are used to provide good
    visual effects for ranging and motion.

    The code is multi-tasking and does the following:
    * Periodically reads button switches SW0 and SW1
    * Display information like PWM values, ranges and WiFi active flag
    * Performs a range of RGB LED animationss
    * Controls a VL53L1X laser distance sensor mounted in the head
    * Three move engines which control the head and leg servos

    The HexBot is controlled via a Wii Nunchuk controller, over a 2.4GHz wireless
    link. After RESET the robot must be brought into the 'Active' state by pressing
    and holding the Nunchuk 'C' button, before any movements can be controlled. The
    following control features are included:

    C(H)      If C is held for >1 second whilst inactive, enter the active state.
    C(H)+Z(H) If C and Z are held for >2 seconds in the active state, enter the
              inactive resting state.
    C         Each C press will increase the responsiveness of the robot from 1–5 (max).
    C(H)+Joy  If C is held in min speed the leg up height will be increased to enable
              it to climb over small objects and uneven surface.
    Z         Each Z press will decrease the responsiveness of the robot from 1–5 (max).
    Z(H)+Joy  Holding Z will change the right/left walking modes from turning to walking
              sideways, forwards will become a ‘bow’ and reverse will become a ‘Bye wave’.
    
    The servo motors must be calibrated during construction, and the values noted
    need to be inserted in the angle definitions included in this code. Functions
    that are commented-out, are not needed but you may find useful to use in your
    code. Enjoy!
*/
// Declare Libraries
#include <Arduino.h>            // needed when using an ESP32 micro
#include <HardwareSerial.h>     // serial library
#include <Wire.h>               // I2C library
#include <FastLED.h>            // Neopixel library
#include "SSD1306Wire.h"        // OLED display library

#include <ESP32Servo.h>         // servo library

#include <esp_now.h>            // ESP-NOW WiFi link library
#include <WiFi.h>               // ESP WiFi library
#include <EEPROM.h>             // EEPROM flash access library
#include "SparkFun_VL53L1X.h"   // VL53L1X laser ranger library

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 21;         // I2C SDA GPIO pin
const int SCL_PIN = 22;         // I2C SCL GPIO pin

// Initialize the oled display for address 0x3c
SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SCL_PIN);

// Define servos constants gained from servo calibration tasks
// All servos are different and leaver arm positions will not match
// These values must be those taken from PWM tests performed on your robot
// If you ever change a servo you will need to re-calibrate those values
#define Ang0_69 2280  // servo 0 front right claw 69°
#define Ang0_90 2086  // servo 0 front right claw 90°
#define Ang0_223 713  // servo 0 front right claw 223°
#define Ang1_21 2034  // servo 1 front right shoulder 21°
#define Ang1_90 1332  // servo 1 front right shoulder 90°
#define Ang1_120 964  // servo 1 front right shoulder 120°
#define Ang2_69 2355  // servo 2 centre right claw 69°
#define Ang2_90 2177  // servo 2 centre right claw 90°
#define Ang2_223 844  // servo 2 centre right claw 223°
#define Ang3_30 2084  // servo 3 centre right shoulder 30°
#define Ang3_90 1490  // servo 3 centre right shoulder 90°
#define Ang3_150 839  // servo 3 centre right shoulder 150°
#define Ang4_68 2007  // servo 4 rear right shoulder 68°
#define Ang4_90 1734  // servo 4 rear right shoulder 90°
#define Ang4_180 830  // servo 4 rear right shoulder 180°
#define Ang5_69 2208  // servo 5 rear right claw 69°
#define Ang5_90 2007  // servo 5 rear right claw 90°
#define Ang5_223 708  // servo 5 rear right claw 223°
#define Ang6_68 1145  // servo 6 rear left shoulder 68°
#define Ang6_90 1394  // servo 6 rear left shoulder 90°
#define Ang6_180 2267 // servo 6 rear left shoulder 180°
#define Ang7_69 514   // servo 7 rear left claw 69°
#define Ang7_90 751   // servo 7 rear left claw 90°
#define Ang7_223 2128 // servo 7 rear left claw 223°
#define Ang8_69 607   // servo 8 centre left claw 69°
#define Ang8_90 784   // servo 8 centre left claw 90°
#define Ang8_223 2150 // servo 8 centre left claw 223°
#define Ang9_30 878   // servo 9 centre left shoulder 30°
#define Ang9_90 1543  // servo 9 centre left shoulder 90°
#define Ang9_150 2103 // servo 0 centre left shoulder 150°
#define AngA_69 611   // servo 10 front left claw 69°
#define AngA_90 788   // servo 10 front left claw 90°
#define AngA_223 2153 // servo 10 front left claw 223°
#define AngB_21 858   // servo 11 front left shoulder 21°
#define AngB_90 1546  // servo 11 front left shoulder 90°
#define AngB_120 1917 // servo 11 front left shoulder 120°
#define ClawRst 75    // resting angle for claws
#define Head_20 560   // fully left head angle, looking at left body corner
#define Head_90 1345  // mid point head angle, looking directly forward
#define Head_160 2130 // fully right head angle, looking at right body corner
#define HeadPin 2     // head servo is connected to GPIO2
#define IO1 15        // small angle offset applied at the start
#define LegDwnMin 110 // min angle for leg down in 'Move Engine' Walk
#define LegDwnMax 120 // max angle for leg down in 'Move Engine' Walk
#define LegRst 90     // resting angle for leg shoulder servos
#define LegUpMin 120  // min angle for leg raised in 'Move Engine' Walk
#define LegUpMax 130  // max angle for leg raised in 'Move Engine' Walk
#define Max0 Ang0_69  // max value for servo 0
#define Max1 Ang1_21  // max value for servo 1
#define Max2 Ang2_69  // max value for servo 2
#define Max3 Ang3_30  // max value for servo 3
#define Max4 Ang4_68  // max value for servo 4
#define Max5 Ang5_69  // max value for servo 5
#define Max6 Ang6_180 // max value for servo 6
#define Max7 Ang7_223 // max value for servo 7
#define Max8 Ang8_223 // max value for servo 8
#define Max9 Ang9_150 // max value for servo 9
#define MaxA AngA_223 // max value for servo 10
#define MaxB AngB_120 // max value for servo 11
#define Min0 Ang0_223 // min value for servo 0
#define Min1 Ang1_120 // min value for servo 1
#define Min2 Ang2_223 // min value for servo 2
#define Min3 Ang3_150 // min value for servo 3
#define Min4 Ang4_180 // min value for servo 4
#define Min5 Ang5_223 // min value for servo 5
#define Min6 Ang6_68  // min value for servo 6
#define Min7 Ang7_69  // min value for servo 7
#define Min8 Ang8_69  // min value for servo 8
#define Min9 Ang9_30  // min value for servo 9
#define MinA AngA_69  // min value for servo 10
#define MinB AngB_21  // min value for servo 11
#define Rst0 Ang0_69 // reset value for servo 0
#define Rst1 Ang1_90 // reset value for servo 1
#define Rst2 Ang2_69 // reset value for servo 2
#define Rst3 Ang3_90 // reset value for servo 3
#define Rst4 Ang4_90 // reset value for servo 4
#define Rst5 Ang5_69 // reset value for servo 5
#define Rst6 Ang6_90 // reset value for servo 6
#define Rst7 Ang7_69 // reset value for servo 7
#define Rst8 Ang8_69 // reset value for servo 8
#define Rst9 Ang9_90 // reset value for servo 9
#define RstA AngA_69 // reset value for servo 10
#define RstB AngB_90 // reset value for servo 11

// Define general constants
#define AngDifMax 250     // maximum sweep angle from centre difference, in microseconds
#define AngDifMin 100     // minimum sweep angle from centre difference, in microseconds
#define autoRestMax 32000 // autoRest max counter value
#define BatCal 383.9      // calibrated voltage multiplying factor
#define BatCritical 2534  // critical battery threshold @6.6v, default = 2672, set == 0 to ignore
#define BatMax 3148       // A0 for fully charged battery voltage, == 8.2v
#define BatPin 36         // analog input pin ADC0 on GPIO36
#define BatWarn 2687      // battery threshold of 7.0v gives low warning
#define DbLLX 117         // JoyX deadband lower limit
#define DbULX 137         // JoyX deadband Upper limit
#define DbLLY 118         // JoyY deadband lower limit
#define DbULY 138         // JoyY deadband Upper limit
#define HeadSpd 24        // speed used in autonomous mode
#define INT_PIN 39        // VL53L1X GPIO interrupt to ADC input not actually used
#define LED_Pin 5         // assign GPIO5 for RGB LEDs
#define LtofDet 350       // max range for target detection in head scanning
#define LtofLimit 500     // Range max value is always limited to this figure
#define LtofMax 400       // out of range value for head scanning
#define LtofMin 55        // minimum range value
#define LtofStop 140      // range value at which autonomous walking stops
#define NumLEDs 13        // number of neopixel LEDs
#define sw0Pin 12         // left switch SW0 assigned to GPIO13
#define sw1Pin 15         // right switch SW1 assigned to GPIO15
#define TESTmode false    // true during assembly, false for normal use
#define WalkSpeedToMax 250; // set WalkSpeed auto-reduction as 5 seconds
#define XSHUT_PIN 14      // GPIO14 controls VL53L1X XSHUT HIGH enable pin
 
// create a FastLED instance and define the total number of LEDs on the robot
CRGB LED[NumLEDs];

// define a data structure matching an LED, used in LED pattern generation
struct LED_rgb {
  uint8_t r;
  uint8_t g;
  uint8_t b;
};
struct LED_rgb LEDX;      // create LEDX as a remporary RGB store

// create servo instances
Servo servoHead;    // define HexBot 2 head servo instance
Servo servoInst[12]; // define HexBot 2 leg servos

// create VL53L1X range sensor instance, and define its pins
// Note that INT_PIN is declared for the library, but not used in the code
SFEVL53L1X VL53L1X(Wire, XSHUT_PIN, INT_PIN);

// my Wii Transciever unique MAC: 50:02:91:68:F7:3F
// you need to build a Wii Transceiver and get its MAC address
uint8_t broadcastAddress[] = {0x50, 0x02, 0x91, 0x68, 0xF7, 0x3F};

// 250 byte max for Tx/Rx data
// It must match the receivers data structure
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;

// Create an 8 byte message called NunchukDataTx, to hold I2C data messages
ESPNOW_Buff250 Tx_Buff;

// Create an 8 byte message to hold incoming sensor readings
ESPNOW_Buff250 Rx_Buff;

// define servo arrays
bool servoAtt[12];// == true when servos are attached
int servoCtr[12]; // centre value sent to servos
// servo          { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11) servo GPIO pin references
int servoPins[] = {19,23,17,18, 4,16,13,27,26,25,33,32};  // servo output pins assigned
int servoMax[12]; // servo max pulse
int servoMem[12]; // record of values sent to servos, used in fidgeting
int servoMin[12]; // servo min pulse
int servoTgt[12]; // target values for servos
int servoVal[12]; // value sent to servos

// Declare and initialise global variables
int16_t AngDif, AngNeg, AngPos, AngRef; // temporary head angles
int16_t Any;                  // any temp value
int16_t AnyCnt;               // general counter
int16_t anyDel;               // any delay value
int16_t anyFor;               // any for loop value
int anyInt;                   // any temporary integer value
long AnyLong;                 // any temp value
unsigned long anyMilli;       // any millisecond timer
int16_t anyVal;               // any temp variable
int16_t autoHeadOFF;          // if set > 0 then counts down to head servo switch OFF, if -1 disabled
int16_t automCnt;             // delay counter used in autonomous mode
int16_t AutoMode;             // autonomouse mode status for display
int16_t automRnd;             // random walking distance in autonomous mode
int16_t automTurn;            // flag used in autonomous turning
int16_t autoOFF;              // if set > 0 then counts down to servo switch OFF, if -1 disabled
int16_t autoRest;             // returns to a rest position after a timeout
String Any$;                  // temporary string
bool atHeadTgt;               // = true if head servo is at target value
bool atTgt;                   // = true if servos are at target values
int32_t BatAvg;               // preload averaging function with high value
int32_t BatDist;              // used battery discharge testing
int32_t BatSum;               // preload averaging sum with 20x max value
int32_t BatVol;               // instantaneous battery voltage
bool Blink;                   // LED blink clock
int16_t BlinkCnt;             // counter for LED Blink clock
bool Border;                  // if true then display has a border
uint8_t Bright;               // FastLED brightness
uint8_t Brightness;           // Display brightness
int CalcCnt;                  // tracks the number of head calculations in autonomous mode
int16_t C_Cnt;                // counter used in 'C' button detection
byte C_Dn;                    // counter for 'C' button down period
byte Col_B,Col_G,Col_R;       // current colours
byte C_Up;                    // counter for 'C' button up period
char cmdMode;                 // command mode
int16_t cmdSgn;               // cmdVal sign, 0 or -1
char cmdType;                 // command mode type
int cmdVal;                   // value associated with a cmdType
int16_t CZ;                   // received button values
bool CZ_Wait;                 // flag used for button release wait function
int16_t DetCnt;               // timer counter used in auto servo detachment
bool DispBack;                // flag used to draw a back arrow on the display
bool DispChng;                // flag invokes initialisation code
bool DispClr;                 // set == true to clear the memory contents
int16_t DispCnt;              // counter used in display updating
int16_t DispDel;              // >0 display delay counter, stops updates over I2C
bool DispDisp;                // set = true to immediately display memory contents
byte DispMenu;                // group type, 0 = sensors, 1 = tasks, etc
int16_t DispMode;             // the type of display for a given mode
bool DispNOP;                 // set == true to stop display I2C updates
int16_t DispNow;              // if > 0 display special modes
int16_t DispPnt;              // points at the current display mode
byte DispCX,DispCY;           // display cursor coordinates
bool ESC;                     // true to escape all functions
bool Echo;                    // if true send print messages
bool HeadActive;              // = true if head is in scan mode for tasks
bool HeadAdj;                 // flag used in head scanning 
int HeadAng;                  // HeadVal converter to degrees
bool HeadAtt;                 // = true when head servo has been attached
int16_t HeadCnt;              // counter used in head movements
int16_t HeadCyc;              // cycle counter used in head movements
int16_t HeadDir;              // head movement direction, +1 == RH, -1 == LH
long HeadErr;                 // value used in head angle adjustment
bool HeadErrSt;               // trip flag used in head scanning
long HeadErrTrip;             // error tripping value used in head angle adjustment
long HeadErrTrk;              // error tracking value used in head angle adjustment
unsigned long HeadInterval;   // main loop head movement interval in microseconds
int16_t HeadMax;              // maximum servo value
unsigned long HeadMicros;     // main loop head movement timer in microseconds
int16_t HeadMin;              // minimum servo value
int16_t HeadMode;             // pointer used in head scanning
int16_t HeadPause;            // 20ms counter to temporarily pause head movement
int16_t HeadSubTask;          // head movement sub task pointer
int16_t HeadTask;             // head movement task pointer
int16_t HeadTaskLast;         // previous head movement task pointer
int16_t HeadTaskNext;         // head movement next task pointer
bool HeadTaskRun;             // set to true during head tasks
int16_t HeadTgt;              // target head servo angle
int16_t HeadTgtCnt;           // counter used in automatic head servo move to target actions
int HeadVal;                  // current head servo angle in microseconds
bool HexBotActive;            // = true if active, otherwise false
bool JoyActive;               // flag used to prevent stack overflow during moves
int16_t JoyD;                 // direction value 0 - 8
int16_t JoyDL;                // direction 1 counter
int16_t JoyMode;              // direction of travel flag, 0 = stationary, 1 = forward, -1 = backward
int JoySpd;                   // speed vector = max(JoyX,JoyY)
int16_t JoyV;                 // joystick vector  = SQRT(SQ(JX) + SQ(JY))
int16_t JoyX;                 // received value
int JoyXV;                    // converted value 0 - 127
int16_t JoyXC;                // received value offset
int16_t JoyY;                 // received value
int16_t JoyYC;                // received value offset
int JoyYi;                    // tracking filter for received JoyY value
int JoyYV;                    // converted value 0 - 128
boolean JoyYT;                // tracking filter flag; true if not on track
char keyChar;                 // any keyboard character
int16_t keyVal;               // any keyboard value
int16_t LedCnt;               // counter used in LED sequencing
int16_t LED_Del;              // delay timer used in LED functions
int16_t LED_Max;              // max value used in LED functions
int16_t LedMode;              // a value > 0 determines LED flash sequence; if = 0 disabled
unsigned long LED_Period;     // LED timer period, default = 20ms
bool LED_Run;                 // == true if LED_Task is to run
bool LEDshow;                 // LED show flag
int16_t LED_Task;             // LED task pointer
bool LedTick;                 // set true every 'Walk' cycle
unsigned long LED_Timer;      // 20+ms LED timer
byte LEDVal;                  // values used in LED tasks
int16_t LegDn;                // dynamic angle for leg down in 'Move Engine'
int16_t LegUp;                // dynamic angle for leg raised in 'Move Engine'
int16_t LegUpFwd;             // dynamic angle for forward leg raised in 'Move Engine'
bool LTOFcalc;                // if == true then calculate speed of ranged object in m/s
int16_t LTOFfps;              // ranger sensor measurement rate in fps
int16_t LTOFperiod;           // range sensor inter-measurementg period in millis()
int16_t LTOFspd10;            // speed of ranged target in m/s x 10
int16_t LTOFspd10L;           // previous speed of ranged target in m/s x 10
int16_t LTOFspdCnt;           // speed peak detector sustain limit
int16_t LTOFt0,LTOFt1;        // range sensor timers for fps
bool MainRun;                 // set true when a MainTask is running to prevent re-entry
int16_t MainTask;             // main task pointer
int16_t MainTaskNext;         // the task to run after an arrest; default is 0
int16_t MoveState;            // move state machine pointer
String myMAC;                 // MAC address if this device
unsigned long next10ms;       // 10ms loop timer
unsigned long next20ms;       // 20ms loop timer
unsigned long next40ms;       // 40ms loop timer
int16_t overRun;              // counter used to control undemanded movement
int16_t Ping;                 // Rx 'ping' counter
unsigned long print40ms;      // 40ms variable print timer
int16_t PrintTgt;             // 0 == serial port, 1 == WiFi ESP-NOW
String PrintTx = "";          // printed strings
int16_t PWM_Freq;             // servo PWM frequency, default = 100;
int Range;                    // range value limited to RangeMax
int RangeRaw;                 // range value obtained from VL53L1X
int16_t RangeAng;             // angle at which minimum range was measured
int16_t RangeCentre;          // distance value looking forward
int RangeDiff;                // difference in RangeRaw from RangeLast
bool RangeEn;                 // == true if VL53L1X ranging is active
bool RangePnt;                // true when a centre range value is recorded, or extremes of sweep
int16_t RangeCentreTracker;   // slugged distance value looking forward
long RangeIntL;               // L-H integrator used in head centring
long RangeIntR;               // R-H integrator used in head centring
int RangeLast;                // previous range measurement
int16_t RangeLeft;            // distance value lookin left
bool RangeLL;                 // == true if range < RangeMin
int RangeMax;                 // the current limited to Range measurements
int RangeMin;                 // the current lower limit of Range measurements
unsigned long RangeMinVal;    // min range value during close range scanning 
int16_t RangeMinAng;          // angle at which RangeMin occured
int16_t RangeMinMem;          // RangeMin at end of each half sweep
int16_t RangeRight;           // distance value lookin right
int16_t RangeTrig;            // range trigger point used in head scanning
bool RangeUL;                 // == true if range exceeds RangeMax
int16_t R0,R1,R2,R3,R4,R5,R6,RM; // range values for backaway and target tracking
bool readWiiCall;             // flag prevents call stack overflow
int16_t Rx_Chk;               // Rx checksum
int16_t RxCZ;                 // buffered received CZ value
int16_t RxJoyX;               // buffered received JoyX value
int16_t RxJoyY;               // buffered received JoyY value
int16_t Rx_len;               // length of WiFi received data block
int16_t Rx_Pnt;               // data pointer user by Rx state machine
bool RxRec;                   // set true when a valid frame of data is received
int16_t RxState;              // receiver state machine state
int16_t Rx_Task;              // task pointer user by Rx state machine
int16_t RTxTimeout;           // counter used to auto-reset if WiFi fails
int16_t RxVal;                // value received from serial Rx
byte RxWiFi[6];               // array holding 6 bytes of Wii received data
bool SerialRx;                // == true if any character has been received
int16_t ServoAny;             // general servo value
int16_t servoDelta;           // temporary delta value used in target movements
bool ServoEn;                 // false = OFF, true = ON enabled status
bool servoHeadEn;             // set = true if head servo is attached
int16_t ServoInc;             // servo angle increment pointer
int16_t ServoPnt;             // servo target pointer for manual control
int16_t SubTask;              // sub task pointer
int16_t SubVal;               // anyvalue used in a subroutine
int16_t sw0Cnt;               // button switch counter
int16_t sw0DwnTime;           // button pressed down time
bool sw0LastState;            // previous state of button switch, HIGH/LOW
bool sw0_Nop;                 // == true to block SW0 read tasks, default = false
bool sw0State;                // state of read button switch pin
int16_t sw0Timer;             // timer used to detemine button sequences
int16_t sw1Cnt;               // button switch counter
int16_t sw1DwnTime;           // button pressed down time
bool sw1LastState;            // previous state of button switch, HIGH/LOW
bool sw1_Nop;                 // == true to block SW1 read tasks, default = false
bool sw1State;                // state of read button switch pin
int16_t sw1Timer;             // timer used to detemine button sequences
unsigned long t0;             // general timer
unsigned long t1;             // general timer
int16_t Task10ms;             // 10ms subtask timer
int16_t Task20ms;             // 20ms subtask timer
int16_t Task40ms;             // 40ms subtask timer
bool TEST;                    // normally false, set = true for assembly test mode
byte Testmode;                // defines the state of TEST
int16_t TestNext;             // task pointer used by the TEST mode
int16_t TestTask;             // task pointer used by the TEST mode
int16_t TgtCnt;               // counter used in automatic servo move to target actions
bool TrgtDet;                 // true if a target has been detected in autonomous mode
bool Turn;                    // default = false; if true turns continuously until = false
bool Turning;                 // set = true so that ESC can be used to break out of a turn cycle
int16_t Tx_PingCnt;           // 1 second Tx ping timer
bool USB;                     // == true if initial battery is below 5.5v
uint8_t VL53L1X_OC;           // VL53L1X ROI SPAD centre
uint16_t VL53L1X_ROIX;        // VL53L1X ROI SPAD width
uint16_t VL53L1X_ROIY;        // VL53L1X ROI SPAD height
unsigned long VL53L1X_Skip;   // read function skip timer
unsigned long VL53L1X_SkipDel;// read function skip timer delay tracker
int16_t VL53L1X_Task;         // if > 0 then run VL53L1X tasks
unsigned long VL53L1X_Timer;  // general timer
unsigned long walkInterval;   // main loop walking interval in microseconds
int16_t Walk;                 // if >0 then walk in one of four directions
long WalkCnt;                 // a counter which controls the walking process
long Walked;                  // counter used to return robot to a given position
unsigned long WalkFreq;       // 1/WalkInterval
int16_t Walk_L;               // walk cmd sweep angle lower limit
int16_t WalkLast;             // previousl Walk value if any
int WalkLftDrive;             // drive factor for left-hand side, 128 = max, 0 = min
byte WalkMax;                 // speed factor 4 - 15
unsigned long walkMicros;     // main loop walking task interval timer in microseconds
int16_t WalkNext;             // used to store the next walking mode
bool WalkPhase;               // walk cycle phase
int16_t Walk_S;               // walk cmd  start angle/count
int16_t WalkSpeed;            // value used to determine WalkMax
int16_t WalkSpeedTO;          // Time-Out counter for WalkSpeed auto-reduction, counts to zero
int WalkRgtDrive;             // drive factor for right-hand side, 128 = max, 0 = min
int16_t Walk_U;               // walk cmd sweep angle upper limit
int W_Dn;                     // leg down angle for walk engine
int16_t WiFiCntC;             // C button counter for WiFi enable
int16_t WiFiConCnt;           // counter used in connection retry times
bool WiFiConnected;           // == true if a WiFi connection has been established
bool WiFiEn;                  // true when WiFi is enabled with prolonged 'Z' button activity
int16_t WiFiPing;             // Rx counter used to maintain connected state
int16_t WiFiRestCnt;          // counter used to goi to rest if no WiFi demands
bool WiFiRx;                  // true when a block of data has been received
long WiFiRxBytes;             // number of bytes received over the WiFi link
int16_t WiFiRxErr;            // number of bytes received over the WiFi link
bool WiFiRxRec;               // true when a block of data has been successfully received
int16_t WiFiTryCnt;           // WiFi try to connect count on error count
int16_t WiFiTryNum;           // WiFi try to connect total count
bool WiFiTx;                  // true when a block of data has been sent successfully
int16_t WiFiTx_CB;            // == 1 call back is clear to send, == -1 resend
int16_t WiFiTx_Chk;           // Tx checksum
int16_t WiFiTx_len;           // >0 when WiFi Tx buffer contains data to be sent
long WiFiTxBytes;             // number of bytes received over the WiFi link
int16_t WiFiTxErr;            // WiFi data error
int16_t WiFiTxErrCnt;         // WiFi Tx error count, leads to disconnection
int16_t WiiType;              // device type, 0 == Nunchuk, 1 == Classic
int WL_C, WR_C;               // walk engine counters
int WL_CM, WR_CM;             // walk engine counters modified by steering
int WL_I, WR_I;               // walk engine counter increments
int W_Up;                     // leg up angle for walk engine
byte Z_Dn;                    // button pressed state; 0 == UP, >= 1 == Down

// ---------------------------------------------------------------------------------

void setup() {
  // Setup code, runs once, immediately after reset
  // Allow allocation of all timers
  ESP32PWM::allocateTimer(0);
  ESP32PWM::allocateTimer(1);
  ESP32PWM::allocateTimer(2);
  ESP32PWM::allocateTimer(3);

  pinMode(sw0Pin,INPUT_PULLUP); // button switch sw0 Pin
  pinMode(sw1Pin,INPUT_PULLUP); // button switch sw1 Pin
  pinMode(BatPin,INPUT);        // ADC0  battery divider
  pinMode(XSHUT_PIN,OUTPUT); digitalWrite(XSHUT_PIN,LOW);     // switch VL53L1X OFF
  pinMode(LED_BUILTIN,OUTPUT); digitalWrite(LED_BUILTIN,LOW); // switch blue LED ON and pull down Servo 12 signal

  setDefaults();        // reset variables

  Wire.begin();         // initialise the I2C interface

  Serial.begin(115200); // baud rate for serial monitor USB Rx/Tx comms

  // initialise the FastLED component
  FastLED.addLeds<WS2812B, LED_Pin, GRB>(LED, NumLEDs);
  FastLED.setBrightness(Bright);
    
  // initialise WiFi link
  WiFi.mode(WIFI_STA);
  myMAC = WiFi.macAddress();
  Serial.println("my MAC= " + String(myMAC));
  WiFi.disconnect();
  
  // Set ESP-NOW Role as bidirectional
//  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);

  // Init ESP-NOW and returns its status
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);

  // Register peer
  esp_now_peer_info_t peerInfo; //initialize and assign the peer information as a pointer to an addres
  memcpy(peerInfo.peer_addr, broadcastAddress, 6); //copy the value of  broadcastAddress with 6 bytes to peerInfo.peer_addr
  peerInfo.channel = 0;  //channel at which the esp talk. 0 means undefined and data will be sent on the current channel. 1-14 are valid channels which is the same with the local device 
  peerInfo.encrypt = false; //not encrypted

  //Add the device to the paired device list 
  if (esp_now_add_peer(&peerInfo) != ESP_OK){ 
    Serial.println("Failed to add peer");
    return;
  }

  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
  
  RunPOST();  // do some basic checks
  
  // set loop timers
  synchLoopTimers(0);
}

// ---------------------------------------------------------------------------------

void(* resetFunc) (void) = 0; //declare reset function at address 0

// ---------------------------------------------------------------------------------

void loop() {
  // Main loop code, runs repeatedly.

  //###############################################################################
  //
  //  WiFi comms check & respond
  //
  //###############################################################################
  // if WiFi is conencted then check Rx buffer on every cycle, so that we respond to
  // received packets as soon as possible
  if (WiFiConnected) {
    if (WiFiRx) {readWiFiRx();} // WiFi Rx buffer contains unread data so deal with it before adding new
  }

  readSerial();                 // check serial port on every loop cycle

  if (VL53L1X_Task > 0) {VL53L1X_Run();}  // perform laser measurements
  
  //###############################################################################
  //
  //  Head movement tasks
  //
  //###############################################################################
  // perform these tasks at nominally 20ms intervals
  if (micros() >= HeadMicros) {
    // move the head towards the target if counter is set
    HeadMicros = HeadMicros + HeadInterval;
    HeadMoveEngine(); // perform head movement tasks
  }

  
  //###############################################################################
  //
  //  Leg movement tasks
  //
  //###############################################################################
  // perform these tasks at nominally 20ms intervals
  if (micros() >= walkMicros) {
    // walking uses a separate timer so that the speed of walking can be controlled
    // before walking is enabled ensure that start conditions are set
    walkMicros = walkMicros + walkInterval; // set next time point
    LedTick = true; // perform LED sequences at same rate as walking
//    Serial.print(RangeIntL);  Serial.print("\t");  Serial.print(RangeIntR);  Serial.print("\t");  Serial.print(RangeAng);  Serial.println("\t 0 \t 12000");
//    Serial.print(Head_65P);  Serial.print("\t");  Serial.print(Head_65N);  Serial.print("\t");  Serial.print(RangeAng);  Serial.println("\t 520 \t 1849");
    if (Walk > 0) {
      // currently walking so call the engine function
      Walk_Engine();
    } else if (TgtCnt > 0) {
      // if not walking see if we need to move the servos?
      Leg_Engine();
    } else if (DetCnt > 0) {
      // setting DetCnt > 0 will detach leg servos after leg movements have stopped
      DetCnt--;
      if (DetCnt < 1) {detachLegServos(); detachHeadServo();}
    }
  }
  
  //###############################################################################
  //
  //  10ms Tasks
  //
  //###############################################################################
  // perform these tasks at 10ms intervals
  if (Task10ms > 0) {
    switch(Task10ms) {
      case 1: readSW1(); break;  // read right button switch SW1
      case 2: readSW0(); break;  // read left button switch SW0
    } Task10ms--; // count down the sub-tasks from 3 - 1 as we loop
  } else if (millis() >= next10ms) {
    // do these every 10ms
    next10ms += 10; // set next time point
    Task10ms = 2;   // initiate 10ms task sequence
  }

  //###############################################################################
  //
  //  RGB LED 20+ ms Tasks (25Hz)
  //
  //###############################################################################
  // The LED tasks are run at a variable rate
  if (LED_Run) {LED_Main();}
  if (millis() >= LED_Timer) {
    LED_Timer+= LED_Period;
    if (LEDshow) {FastLED.show(); LEDshow = false;} // update LEDs if LEDshow == true
    LED_Run = true; // initiate the next LED_Task process
  }
  
  //###############################################################################
  //
  //  20ms Tasks
  //
  //###############################################################################
  // perform these tasks at 20ms intervals
  if (Task20ms > 0) {
    switch(Task20ms) {
      case 1: battRead(); break;            // read analogue input
      case 2: if (TEST) {TestMoveServos();} // move servos, if enabled
        break;
      case 3: if (autoHeadOFF > 0) {
          autoHeadOFF--; if (autoHeadOFF < 1) {detachHeadServo();}
        } break;
      case 4: if ((HexBotActive) && (autoRest < autoRestMax)) {CheckAutoRest();}
        break;
      case 5: DoHeadTasks(); break;
    } Task20ms--; // count down the sub-tasks from 5 - 1 as we loop
  } else if (millis() >= next20ms) {
    // do this every 20ms
    next20ms += 20; // set next time point
    Task20ms = 5;   // initiate 20ms task sequence

    // if connected to WiFi read Wii data and respond
    if (WiFiConnected) {readWii();}

    //###############################################################################
    //
    //  Main tasks
    //
    //###############################################################################
    // perform 'Main' tasks in manual button mode
    switch (MainTask) {
      case 0: break; // default null task
      case 1: // goto HexBotActive default standing position
        MainTask_LTOF(); break;
      case 2: // backaway from target mode, then return to start
        MainTask_TrackWall(0); break;
      case 3: // track wall target mode
        MainTask_TrackWall(1); break;
      case 4: // autonomous demo mode
        MainTask_Autonomous(); break;
    }

    // perform auto-speed reduction if needed, applies to speeds 4 & 5 only
    // this is aimed at preventing servo burn-out, limiting high speed bursts
    if (WalkSpeedTO > 0) {
      WalkSpeedTO--; if (WalkSpeedTO == 0) {
        if (WalkSpeed == 4) {WalkSpeed = 3; WalkMax = 10; WalkSpeedTO = 2 * WalkSpeedToMax;}
        if (WalkSpeed == 5) {WalkSpeed = 4; WalkMax = 13; WalkSpeedTO = WalkSpeedToMax;}
      }
    }

    // if WiFi fails stop movement and trigger autoRest
    // if readWiiCall == true then we are currently locked in a movement, so ignore timeout
    if (HexBotActive && (RTxTimeout > 0) & !readWiiCall) {
      RTxTimeout--;
      if (RTxTimeout == 1) {
        Serial.println("WiFi failed!...");
        Walk = 0; TgtCnt = 0; LedMode = 0; autoRest = 0;
      }
    }
    
    // if not moving do the following:
    if ((Walk < 1) && (TgtCnt < 1)) {
      // return to Power-OFF state?
      if (autoOFF > 0) {autoOFF--; if (autoOFF < 1) {
        //detachServos();
        }
      }
    }
  }

  //###############################################################################
  //
  //  40 ms Tasks (25Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 40ms period
  if (Task40ms > 0) {
    switch(Task40ms) {
      case  1: Display_40ms(); break;  // update display if necessary
      case  2: if (Ping > 0) {Ping--;} break;  // time-out data received flag
      case  3: // if not connected to WiFi run then try to connect every second
        if (!WiFiConnected) {
          // only try to connect WiFi if no 'Pings' are being received
          if (Ping == 0) {
            WiFiConCnt--; if (WiFiConCnt < 1) {
              WiFiConCnt = 50; WiFiTryToConnect();
//              Serial.println("Trying to connect WiFi");
            }
          }
        } break;
      case  4:
        if (WiFiPing > 0) {
          WiFiPing--;
          if (WiFiPing < 1) {
             // we have stopped receiving so disconnect WiFi
            WiFiDisconnect();
//            Serial.println("WiFiPing == 0");
          }
        } break;
    } Task40ms--;
  } else if (millis() >= next40ms) {
    next40ms += 40; Task40ms = 4;
  }


  //###############################################################################
  //
  //  40 ms Print Task - Variable
  //
  //###############################################################################
  // this code uses either the serial port or the WiFi link to send messages
  if (millis() >= print40ms) {
    print40ms = millis() + 40;
    if (!Echo) {PrintTx = "";}  // Echo is false so empty print buffer
    if (PrintTx.length() > 0) {
      // characters in string buffer so send some/all of them
      print40ms = millis() + 4; // shorten the delay for more data to send
      Tx_PingCnt = 0;           // reset the WiFi ping counter
      if (PrintTgt == 0) {
        // default uses the serial port
        if (Serial.availableForWrite() >= 64) {
          if (PrintTx.length() <= 64) {
            Serial.print(PrintTx);
            PrintTx = "";   // empty the buffer
          } else {
            Serial.print(PrintTx.substring(0,64));
            PrintTx = PrintTx.substring(64);
          }
        }
      } else {
        // text data has been received from WiFi so respond using ESP_NOW
        // ensure that we have had a call back from the previous frame
        if (WiFiConnected) {
          // only trannsmit PrintTx data if connected
          if (WiFiTx_CB == 1) {
            WiFiTx_CB = 0;  // clear the call back flag when sending
            if (PrintTx.length() <= 32) {
              WiFiTx_len = PrintTx.length();
              PrintTx.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = "";   // empty the buffer
            } else {
              WiFiTx_len = 32;
              Any$ = PrintTx.substring(0,WiFiTx_len);
              Any$.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = PrintTx.substring(32);
            }
          } else if (WiFiTx_CB == -1) {
            // an error has occured so resend the data
            if (WiFiTryCnt < 5) {
              WiFiTryCnt++;   // track the number of retries
              WiFiTx_CB = 0;  // clear the call back flag when sending
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
            } else {
              // too many resends so disconnect WiFi link
              WiFiDisconnect();
              Serial.println("Failed to Tx!");
            }
          }
        } else {
          // not connected so clear PrintTxdata
          PrintTx = "";
        }
      }
    }
  }
}

// --------------------------------------------------------------------------------

void delayLoop(unsigned long zDel) {
  // loop for a delay period
  zDel = millis() + zDel;
  while (millis() < zDel) {
    if (ESC) return;
    loop(); yield();
  }  
}

// --------------------------------------------------------------------------------

void loopWhileTgtCnt() {
  // used to call loop() code during a 'Move' function
//  Serial.println(F("loopWhiletgtCnt"));
  while (TgtCnt > 0) {
    if (ESC) {return;}
    loop(); yield();
  }
}

// --------------------------------------------------------------------------------

void loopWhileTurn(long zWalkCnt) {
  // count number of Walked clicks whilst turning, ignoring range data
  // from testing the corection factor for 360° = 2.3
  // zWalkCnt is effective the angle passed to this funtion
  zWalkCnt = (zWalkCnt * 23)/10;
  Walked = 0; // clear the walked counter
  while (zWalkCnt > Walked) {
    if (ESC) return;
    loop();
    // slowly increase walking speed
    if (walkInterval > 15000) {walkInterval = walkInterval -100;}
    yield();
  }
}

// --------------------------------------------------------------------------------

void loopWhileTurnEcho(long zWalkCnt) {
  // count number of Walked clicks whilst turning
  // from testing the corection factor for 360° = 2.3
  // zWalkCnt is effective the angle passed to this funtion
  // if there is no wall/target detected then return by setting ESC
  zWalkCnt = (zWalkCnt * 23)/10;
  Walked = 0; // clear the walked counter
  while (zWalkCnt > Walked) {
    if (ESC) return;
    loop(); yield();
    // check range
    if (RangeMinMem > LtofStop){ESC = true;}
    // slowly increase walking speed
    if (walkInterval > 15000) {walkInterval = walkInterval -100;}
  }
}

// --------------------------------------------------------------------------------

void loopWhileWalkEcho(long zWalkCnt) {
  // walk for a number of Walked clicks whilst reading the radar distance
  // there are nominally 32 clicks in a walk leg stride
//  Serial.println("loopWhileWalk");
  Walked = 0; // clear the walked counter
  while (zWalkCnt > Walked) {
    if (ESC) return;
    loop();
    // check Range reading and exit if an object is detected
    if (Range <= 325) {break;}
    // slowly increase walking speed
    if (walkInterval > 15000) {walkInterval = walkInterval -100;}
    yield();
  }
}

// --------------------------------------------------------------------------------

void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  // Callback when data is received
  // we copy the data into a buffer and set a flag so that the main loop will act
  // to handle it
  while (WiFiRx) {
    // buffer contains unread data so deal with it before adding new
    readWiFiRx();
  }

  memcpy(&Rx_Buff, incomingData, len);

  // incoming data will always be two or more characters, ie. text == $+++
  if (len < 2) {Serial.println("Rx Error - no Rx data"); return;}
  
  if (Rx_len > 0) {Serial.println("Rx Error - buffer in use");} // this should not happen

  // if connected Rx data will have a checksum so check it first
  Rx_len = len - 1; Rx_Pnt = 0;
  RxGetChecksum(Rx_len);  // generate a checksum from Rx data
  if (Rx_Chk != Rx_Buff.ESPdata[len - 1]) {
    Serial.println("Rx Error - checksum");
  }
  // set flags before exiting
  WiFiRxBytes += Rx_len; Rx_Task = 0; WiFiRx = true;
  WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
  WiFiRxRec = true; // set flag indicating data has been recieved from Wii Transceiver
//    Serial.print("Bytes received: "); Serial.println(len);
}

// --------------------------------------------------------------------------------

void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  // Callback when data is sent, determines success or failure
//  t1 = micros();
//  Serial.print("Send Status: ");
  if (WiFiConnected) {
    // the link is already active so handle normal success/failure scenarios
    if (status == 0){
//      Serial.println("Success " + String(t1-t0));
      WiFiTx = true;    // flag WiFi Tx success
      WiFiTx_CB = 1;    // set flag to indicate that a call back state has occured
      WiFiTxErrCnt = 0; // clear the error flag in case of a glitch
      WiFiPing = 50;    // set link time-out to 2 seconds
    } else {
//      Serial.println("Fail " + String(t1 - t0));
      WiFiTx = false;   // flag WiFi Tx failure
      WiFiTx_CB = -1;   // set flag to indicate that a resend is required
      WiFiTxErrCnt++;
    }
  } else {
    // not connected so we must have tried to connect and may get an error
    if (status == 0){
      // successful so enter the connected state
      Serial.println("Connected!");
      WiFiConnected = true;
      WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
      WiFiTxErrCnt = 0; // reset error count for this connection
    } else {
//      Serial.println("Status = " + String(status));
      WiFiTxErrCnt++;
    }
    WiFiTx_CB = 1;
  }
}

// ---------------------------------------------------------------------------------

void RunPOST() {
  // start-up initialisation processes  
  setDefaults();  // reset variables
  Serial.println("\n\nHexBot2 v0.0\n");

  // initilaise RGB LED array
  FastLED.clear();
  FastLED.show(); // This sends the updated pixel color to the hardware.
  Serial.println("RGB LEDs cleared.");
  LED_Set_All(0,1,0); // start all green
  
  // Initialise display
  Display_Init();
  display.setBrightness(Brightness);
  Display_Intro();
  Serial.println("Display initialised");
  DispDel = 50; // wait 2 seconds before displaying anything else

  // check for SW0 button being pressed when coming out of REST
  if (digitalRead(sw0Pin) == LOW) {
    // SW0 is being pressed so force TEST mode
    TEST = true;
  }

  // initialise VL53L1X sensor
  VL53L1X_ON();   // turn it ON briefly to confirm that it is there
  VL53L1X_OFF();  // turn the laser OFF by default

  // set up battery quick averaging
  BatVol = analogRead(BatPin);
  BatAvg = BatVol; BatSum = BatVol * 50;  // preload the battery sum for averaging
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
  Serial.println("Battery " + String(BatVol) + "\t" + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");
  randomSeed(BatVol); // use the battery voltage as the random seed factor
  // if battery is less than 5.5v then USB mode is assumed, which makes servos inoperable
  if (BatAvg <= 2111) {USB = true;}

  // get ready to run
  if (TEST) {
    // in TEST mode, used during assembly
    // LEDs cycle RGB colours
    // Display has a border and cycles through settings
    // SW0 sets servos to 1500 constant or short sweep
    // SW1 toggle LTOF sensor ON/OFF
    Serial.println("Running in TEST mode");
    delay(1000); //display intro for 1 second
    Display_TestMode();
    DispDel = 50;       // wait 2 seconds before displaying anything else
    for (int zS = 0; zS <= 12; zS++) {
      servoVal[zS] = 1500;  // set default servo values
      servoAtt[zS] = false; // set 'attached' flag
    }
    HeadVal = 1500; // correct target positions
    HeadAtt = false;
    Wait_ButtonsDwn();  // ensure no buttons are still pressed
  } else {
    // normal mode so display battery voltage after intro screen
    Serial.println("Running in normal mode");
    delay(2000);
    LED_Set_All(1,0,0); // all red as warning
    HeadToRest();               // centre head servo
    HeadMoveNow();              // complete the movement
    delay(2000); LED_Del = 50;
    // now progressively move legs to store positions, starting with rear shoulders
    SetAngle(24,160); servoVal[4] = servoTgt[4]; attachAServo(4);
    SetAngle(26,160); servoVal[6] = servoTgt[6]; attachAServo(6);
    delay(100);
    // next is centre shoulders
    SetAngle(23,140); servoVal[3] = servoTgt[3]; attachAServo(3);
    SetAngle(29,140); servoVal[9] = servoTgt[9]; attachAServo(9);
    delay(100);
    // next is front shoulders
    SetAngle(21,100); servoVal[1] = servoTgt[1]; attachAServo(1);
    SetAngle(31,100); servoVal[11] = servoTgt[11]; attachAServo(11);
    delay(100);
    // now attach the rest and move to store position
    GoToStore(50); autoOFF = 5; // move legs to REST position
   }
  Serial.println("POST completed");
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values
  AnyCnt = 0;             // general counter
  autoHeadOFF = 0;        // if set > 0 then counts down to head servo switch OFF, if -1 disabled
  AutoMode = 0;           // autonomouse mode status for display
  autoOFF = 0;            // if set > 0 then counts down to servo switch OFF, if -1 disabled
  autoRest = autoRestMax; // returns to a rest position after a timeout
  atHeadTgt = false;      // = true if head servo is at target value
  atTgt = false;          // = true if servos are at target values
  BatAvg = 4095;          // preload averaging function with high value
  BatAvg = 0;             // preload averaging function with high value
  BatSum = 72320;         // preload averaging sum with 20x max value
  Blink = false;          // LED blink clock
  BlinkCnt = 0;           // counter for LED Blink clock
  Border = true;          // if true then display has a border
  Bright = 255;           // FastLED brightness
  Brightness = 128;       // Display brightness
  C_Cnt = 0;              // counter used in 'C' button detection
  C_Dn = 0;               // counter for 'C' button down period
  cmdMode = ' ';          // command mode
  cmdSgn = 1;             // cmdVal sign, 1 or -1
  cmdType = ' ';          // command mode type
  cmdVal = 0;             // value associated with a cmdType
  CZ = 0;                 // received button values
  CZ_Wait = false;        // flag used for button release wait function
  DetCnt = 0;             // timer counter used in auto servo detachment
  DispBack = false;       // flag used to draw a back arrow on the display
  DispChng = false;       // flag invokes initialisation code
  DispClr = false;        // set == true to clear the memory contents
  DispCnt = 0;            // counter used in display updating
  DispDel = 0;            // >0 display delay counter, stops updates over I2C
  DispDisp = false;       // set = true to immediately display memory contents
  DispMenu = 0;           // group type, 0 = sensors, 1 = tasks, etc
  DispMode = 0;           // the type of display for a given mode
  DispNOP = false;        // set == true to stop display I2C updates
  DispNow = 0;            // if > 0 display special modes
  DispPnt = 0;            // points at the current display mode, default = 0 batteries
  Echo = true;            // if true send print messages
  ESC = false;            // set = true to escape all functions
  HeadActive = false;     // true if head is in scan mode for tasks
  HeadAtt = false;        // = true when head servo has been attached
  HeadDir = 0;            // head movement direction, +1 == RH, -1 == LH
  HeadInterval = 20000;   // main loop head movement interval in microseconds
  HeadPause = 0;          // 20ms counter to temporarily pause head movement
  HeadSubTask = 0;        // head movement sub task pointer
  HeadTask = 0;           // head movement task pointer
  HeadTaskRun = false;    // set to true during head tasks
  HeadTgtCnt = 0;         // counter used in automatic head servo move to target actions
  HeadVal = Head_90 + 200; // current head servo angle, set to 200 > Head_90
  HeadTgt = HeadVal;      // target head servo angle
  HexBotActive = false;   // = true if active, otherwise false
  JoyActive = true;      // flag used to prevent stack overflow during moves
  JoyDL = 0;              // direction 1 counter
  JoyMode = 0;            // direction of travel flag, 0 = stationary, 1 = forward, -1 = backward
  JoySpd = 0;             // speed vector = max(JoyX,JoyY)
  JoyX = 128;             // received value, default
  JoyXV = 0;              // converted value 0 - 127
  JoyY = 128;             // received value, default
  JoyYi = 0;              // tracking filter for received JoyY value
  JoyYV = 0;              // converted value 0 - 128
  LED_Del = 0;            // delay timer used in LED functions
  LED_Period = 20;        // LED timer period, default = 20ms
  LED_Run = false;        // == true if LED_Task is to run
  LEDshow = false;        // LED show flag for mouth LEDs
  LED_Task = 0;           // LED task pointer
  LedTick = false;        // set true every 'Walk' cycle
  LEDVal = 0;             // values used in LED tasks
  LegDn = LegDwnMin;      // dynamic angle for leg down in 'Move Engine'
  LegUp = LegUpMin;       // dynamic angle for leg raised in 'Move Engine'
  LegUpFwd = LegUpMin;    // dynamic angle for forward leg raised in 'Move Engine'
  LTOFcalc = false;       // if == true then calculate speed of ranged object in m/s
  LTOFspd10 = 0;          // speed of ranged target in m/s x 10
  LTOFspd10L = 0;         // previous speed of ranged target in m/s x 10
  LTOFspdCnt = 0;         // speed peak detector sustain limit
  MainRun = false;        // set true when a MainTask is running to prevent re-entry
  MainTask = 0;           // main task pointer
  MainTaskNext = 0;       // the task to run after an arrest; default is 0
  MoveState = 0;          // move state machine pointer
  overRun = 0;            // counter used to prevent undemanded movement
  Ping = 0;               // 'ping' counter
  PWM_Freq = 100;         // servo PWM frequency, default = 100;
  RangeEn = false;        // == true if VL53L1X ranging is active
  RangeMax = LtofLimit;   // the current limited to Range measurements
  RangeMin = LtofMin;     // the current lower limited to Range measurements
  readWiiCall = false;    // flag prevents call stack overflow
  Rx_Chk = 0;             // Rx checksum
  RxCZ = 0;               // buffered received CZ value
  RxJoyX = 0;             // buffered received JoyX value
  RxJoyY = 0;             // buffered received JoyY value
  Rx_len = 0;             // length of WiFi received data block
  RxRec = false ;         // set true when a valid frame of data is received
  RxState = 0;            // receiver state machine state
  Rx_Task = 0;            // pointer user by Rx state machine
  RTxTimeout = 0;         // counter used to auto-reset if WiFi fails
  RxVal = -1;             // value received from serial Rx
  SerialRx = false;       // == true if any character has been received
  ServoAny = 1500;        // general servo value
  ServoEn = false;        // false = OFF, true = ON enabled status
  servoHeadEn = false;    // set = true if head servo is attached
  ServoInc = 0;           // servo angle increment pointer
  ServoPnt = 0;           // servo target pointer for manual control, default = 0
  
  // servo start values are offset by IO1 to ensure some movement occurs at reset
  SetServoVal( 0,ClawRst + IO1); // starting value of servo
  SetServoVal( 1,    120 - IO1); // starting value of servo
  SetServoVal( 2,ClawRst + IO1); // starting value of servo
  SetServoVal( 3,    150 - IO1); // starting value of servo
  SetServoVal( 4,    150 - IO1); // starting value of servo
  SetServoVal( 5,ClawRst + IO1); // starting value of servo
  SetServoVal( 6,    150 - IO1); // starting value of servo
  SetServoVal( 7,ClawRst + IO1); // starting value of servo
  SetServoVal( 8,ClawRst + IO1); // starting value of servo
  SetServoVal( 9,    150 - IO1); // starting value of servo
  SetServoVal(10,ClawRst + IO1); // starting value of servo
  SetServoVal(11,    120 - IO1); // starting value of servo

  SubTask = 0;            // sub task pointer
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw0_Nop = false;        // == true to block SW0 read tasks, default = false
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw1Cnt = 0;             // button switch counter
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw1_Nop = false;        // == true to block SW1 read tasks, default = false
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  Task10ms = 0;           // 10ms subtask timer
  Task40ms = 0;           // 40ms subtask timer
  TEST = TESTmode;        // normally false, set = true for assembly test mode
  TestTask = 1;           // task pointer used by the TEST mode
  TgtCnt = 0;             // counter used in automatic servo move to target actions
  Turn = false;           // default = false; if true turns continuously until = false
  Tx_PingCnt = 0;         // 1 second Tx ping timer
  USB = false;            // == true if initial battery is below 5.5v
  VL53L1X_OC = 199;       // VL53L1X ROI SPAD centre
  VL53L1X_ROIX = 16;      // VL53L1X ROI SPAD width
  VL53L1X_ROIY = 16;      // VL53L1X ROI SPAD height
  VL53L1X_SkipDel = 31;   // read function skip timer delay tracker, nominally 32
  VL53L1X_Task = 0;       // if > 0 then run LV53L1X tasks
  Walk = 0;               // if >0 then walk in one of four directions
  WalkCnt = 0;            // a counter which controls the walking process
  Walked = 0;             // counter used to return robot to a given position
  walkInterval = 20000;   // main loop walking interval in microseconds
  Walk_L = 85;            // walk cmd sweep angle lower limit
  WalkLast = 0;           // previousl Walk value if any
  WalkLftDrive = 128;     // drive factor for left-hand side, 128 = max, 0 = min
  WalkMax = 16;           // speed factor 16 - 60
  WalkRgtDrive = 128;     // drive factor for right-hand side, 128 = max, 0 = min
  Walk_S = 90;            // walk cmd  start angle/count
  WalkSpeed = 1;          // value used to determine WalkMax, 1 - 5
  WalkSpeedTO = 0;        // Time-Out counter for WalkSpeed auto-reduction, counts to zero
  Walk_U = 95;            // walk cmd sweep angle upper limit
  WiFiCntC = 0;           // C button counter for WiFi enable
  WiFiConCnt = 0;         // counter used in connection retry times
  WiFiConnected = false;  // == true if a WiFi connection has been established
  WiFiEn = false;         // true when WiFi is enabled with prolonged 'Z' button activity
  WiFiPing = 0;           // Rx counter used to maintain connected state
  WiFiRx = false;         // true when a block of data has been received
  WiFiRxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiRxErr;              // number of receive errors in the WiFi link
  WiFiRxRec = false;      // true when a block of data has been successfully received
  WiFiTryCnt = 0;         // WiFi try to connect count
  WiFiTryNum = 0;         // WiFi try to connect total count
  WiFiTx = false;         // true when a block of data has been sent successfully
  WiFiTxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiTx_CB = 1;          // == 1 call back is clear to send, == -1 resend
  WiFiTx_Chk = 0;         // Tx checksum
  WiFiTx_len = 0;         // >0 when WiFi Tx buffer contains data to be sent
  WiFiTxErr = 0;          // WiFi data error
  WiFiTxErrCnt = 0;       // WiFi Tx error count, leads to disconnection
  WiiType = 0;            // device type, 0 == Nunchuk, 1 == Classic
  Z_Dn = 0;               // button pressed state; 0 = UP, 1 = Down
}

// --------------------------------------------------------------------------------

void synchLoopTimers(unsigned long zAdvance) {
  // reset or adjust main loop timers, based on zAdvance
  // if  0 then a 2ms phase offset is introduced to prevent simultaneous triggering
  // if >0 then add zAdvance to the current timer values
  if (zAdvance == 0) {
    // setting up fresh loop timers
    HeadMicros = micros() + HeadInterval; // set head loop timer
    LED_Timer = millis() + 26;  // reset 20+ms loop timer with 3ms phase to 10ms timer
    next10ms = millis() + 10;   // reset 10ms loop timer with 2ms phase to 10ms timer
    next20ms = millis() + 22;   // reset 20ms loop timer with 2ms phase to 10ms timer
    next40ms = millis() + 44;   // reset 40ms loop timer with 2ms phase to 10ms timer
    print40ms = millis() + 46;  // reset 40ms printer loop timer with 4ms phase to 10ms timer
    walkMicros = micros() + walkInterval; // set walk loop timer
  } else {
    // advancing loop timers
    HeadMicros += (zAdvance * 1000);
    LED_Timer += zAdvance;
    next10ms += zAdvance;
    next20ms += zAdvance;
    next40ms += zAdvance;
    print40ms += zAdvance;
    walkMicros += (zAdvance * 1000);
  }
}

// ---------------------------------------------------------------------------------
