// ################################################################################
//
//  Wii WEMOS Transciever v0.01
//
//  Released:  14/10/2020
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    This code is for a 2.4 GHz WiFi transceiver devices that enables a Wii controller
    connected via I2C to interface with remote robot projects.

    Micro:  WEMOS D1 Mini

    MAC:  50:02:91:68:F7:3F

    The following features are supported:
    
    * detection of Wii controller devices; Nunchuk, Classic, etc on I2C
    * reads Wii over I2C every 40ms and transmits data if present
    * battery health monitoring, with auto shut-down
    * WiFi comms based on ESP-Now peer to peer packets
    * several display modes
    * soft Reboot if button held down
    * waits for remote device to initiate a connection
*/
// Declare Libraries
#include <Arduino.h>
#include <EEPROM.h>             // ESP8266 flash emulator
#include <Wire.h>               // library needed to run the I2C bus interface
#include <ESP8266WiFi.h>        // ESP8266 WiFi library
#include <espnow.h>             // ESPNOW peer-to-peer library
#include "SSD1306Wire.h"        // OLED display library

// remote device MAC address is nulled to start with
uint8_t broadcastAddress[] = {0,0,0,0,0,0};

// 250 byte max for Tx/Rx data
// It must match the receivers data structure
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;

// Create an 8Byte_message called NunchukDataTx to hold I2C data message
ESPNOW_Buff250 Tx_Buff;

// Create an 8Byte_message to hold incoming sensor readings
ESPNOW_Buff250 Rx_Buff;

// Define constants
#define BatCal 103.0    // calibrated voltage multiplying factor, default 102.3 
#define BatCritical 700 // critical battery threshold, default = 700, set == 0 to ignore
#define BatMax 978      // fully charged battery voltage, == 9.5v
#define BatUSB 100      // if <= then USB supply assumed
#define BatVolPin A0    // analog input pin connected to battery voltage divider
#define BatWarning 721  // warning battery voltage threshold == 7.0v
#define DbHalf 8        // value used as half of the joystick deadbands
#define WiiErrRet 19    // error threshold used in I2C initialisation

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = D2;
const int SDC_PIN = D1;

// Initialize the oled display for address 0x3c
 SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SDC_PIN);
 
// Define Wemos D1 pcb pin to GPIO mapping
#define D0 16 // GPIO16
#define D1  5 // GPIO5
#define D2  4 // GPIO4
#define D3  0 // GPIO0
#define D4  2 // GPIO2
#define D5 14 // GPIO14
#define D6 12 // GPIO12
#define D7 13 // GPIO13
#define D8 15 // GPIO15

// define EEPROM keys
#define EEPROM_K0Def B10101010  // EEPROM 1st key
#define EEPROM_K1Def B01010101  // EEPROM 2nd key

// Declare and initialise global variables
long AnyLong;             // any temp value
int BatAvg;               // preload averaging function with high value
byte BatCnt;              // pointer to array values
int BatDel;               // set a 10 second delay in 'life' prediction
int BatLast;              // previous battery average reading
bool BatPwr;              // true if battery supply detected, false if USB assumed
int BatQ;                 // a measure of battery quality 1 - 5
int BatSum;               // preload averaging sum with 20x max value
unsigned long BatT0;      // time of first battery measurement in life prediction forecast
String BatTime$;          // time left displayed
unsigned long BatTms;     // time when battery reading was taken
int BatV0;                // reference voltage used in prediction forecast
int BatV1;                // lowest voltage used in prediction forecast
int BatVol;               // instantaneous battery voltage
bool Connected;           // == true if a WiFi connection has been established
int DataTxLen;            // length of data frame to send, min = 1, max = 250
int DataRxLen;            // length of data frame received, min = 1, max = 250
int Dev_ID;               // remote device ID, -1 default unknown
int DispCnt;              // counter used in display updating
bool DispClr;             // set == true to clear the memory contents
int DispDel;              // display hold off counter, default = 0
bool DispDisp;            // set = true to immediately display memory contents
bool DispHold;            // set == true to stop display cycling
int DispNxt;              // display switch timer, default = 50, 2 sec
byte DispPnt;             // points at the current display mode
byte DispCX,DispCY;       // display cursor coordinates
byte JoyXOff;             // value to correct stick centre X offset
byte JoyYOff;             // value to correct stick centre Y offset
String myMAC;             // MAC address if this device
unsigned long next2ms;    // 2ms loop timer trigger for serial data
unsigned long next10ms;   // 10ms loop timer trigger
unsigned long next40ms;   // 40ms loop timer trigger
int nunchuk_address;      // Nunchuk I2C address (0x52)
bool Reboot;              // set true when reboot is in progress to stop all tasks
byte received_data[16];   // Array storing data received from Nunchuk
int reportEn;             // reports data for testing if >= 1. Set this to 0 for transparent telemetry mode
int Rx_Chk;               // Rx checksum
int Rx_len;               // >0 when a packet has been received
int Rx_Pnt;               // receive pointer for data management
byte send_byte;           // bit pattern sent to BalanceBot over WiFi
bool STOP;                // set true when battery is critical
String success;           // variable to store if sending data was successful
int sw1Cnt;               // button switch counter
boolean sw1LastState;     // previous state of button switch, HIGH/LOW
boolean sw1State;         // state of read button switch pin
int sw1Timer;             // timer used to detemine button sequences
unsigned long t0,t1;      // microseconnd function timers
long t1_t0;               // calc of t1 - t0
long t1t0Cnt;             // number of t1-t0 readings
long t1t0Max;             // max of t1-t0 times
long t1t0Min;             // min of t1-t0 times
long t1t0Sum;             // sum of t1-t0 for averaging
int Task40ms;             // task pointer for 40ms tasks
bool Tx;                  // status of transmission, true == ok, false == failed
int Tx_CB;                // == 1 call back is clear to send, == -1 resend
char Tx_Char;             // serial byte received from usb port, to be sent over WiFi
int Tx_Chk;               // Tx checksum
bool Tx_Echo;             // ==true to echo serial data at port level, default false
int Tx_ErrCnt;            // disconnects if error count is too high
bool Tx_Hd;               // == true if a '$' text header needs to be included in a block
int Tx_len;               // >0 when WiFi Tx buffer contains data to be sent
String WiFi_Dev;          // the name of the remote device establishing the connection
int WiFi_State;           // WiFi connection state machine
int WiiCal;               // counter used in offset calibration
bool  Wii_Connected;      // true if a Wii contoller is detected
int Wii_Cnt;              // counter used in Wii connection code
int WiiErrCnt;            // number of attempts to initialse the Nunchuk
byte WiiError;            // used to detect transmission errors
int WiiTask;              // initialisation task pointer
char WiiType;             // device type, 'N' == Nunchuk, 'C' == Classic
bool Wii_Tx;              // ==true to load I2C data into the Tx buffer
bool XYOffAdj;            // if == true adjust XY centre offsets


void setup() {
  // Setup code, runs once:
  pinMode(D4,INPUT_PULLUP); // define SW1 pin as input
  Serial.begin(115200);             // Start the USB serial port at 115200 bps
  Serial.println(F("\n\n\n\nWii Wemos D1 Transceiver v0.01\n"));

  WiFi.mode(WIFI_STA);
  myMAC = WiFi.macAddress();
  Serial.println("MAC: " + myMAC + "\n");
  WiFi.disconnect();

  Set_Defaults(); // reset all variables
  Flush_Serial(); // empty the incoming serial port
  runPOST();
  delay(1000);  // allow things to settle before turning ON WiFi

  // Set ESP-NOW Role as bidirectional
  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);
  
  // Register peer
//  esp_now_add_peer(broadcastAddress, ESP_NOW_ROLE_COMBO, 1, NULL, 0);
  
  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);

  Sync_Timers();
}

// ---------------------------------------------------------------------------------

void(* resetFunc) (void) = 0; //declare reset function at address 0

// --------------------------------------------------------------------------------

void loop() {
  // Main loop code, runs repeatedly.
  if (STOP) {return;}   // stop everything due to low battery
  if (Reboot) {return;} // stop everything as a reset is pending

  //##############################################################################
  //
  // 10ms tasks
  //
  //##############################################################################
  if (millis() >= next10ms) {
    // 10ms tasks
    next10ms = millis() + 10; // rset loop timer

    readSW1(); if (Reboot) {return;}  // read button switch
  }


  //##############################################################################
  //
  // 40ms tasks
  //
  //##############################################################################
  if (Task40ms > 0) {
    switch(Task40ms) {
      case 1: Display_40ms(); break;  // update display if necessary
      case 2: // read battery voltage and predict capacity
        ReadBattery();
        if (STOP) {return;}  // stop if battery voltage is critical
        break;
      case 3: // handle Wii on I2C
        if (Wii_Connected) {
          // if Wii controller is connected read I2C data and send it over the
          // the WiFi link if a remote device is connected, by setting a flag
          ReadI2CData();
          Wii_Tx = true;
//          ReportNunchukData(); Serial.println("");
        } else {
          // if the Wii is not connected then test for it every 400ms
          WiiInitialiseTask();
          WiiTask++; if (WiiTask >= 10) {
            WiiTask = 0;    // reset Wii initialisation task pointer
            Wii_Tx = true;  // send default data every 400ms
          }
        }
        
        if ((reportEn > 0) && Wii_Connected) {
          // reports the contents of the Wii controller data registers, if a controller
          // is connected to I2C
          reportEn--;
          if (reportEn < 1) {
            reportEn = 5;
            Serial.println(F("\nWii I2C Data:")); // print title
            ReportNunchukData();
          }
        } break;
    } Task40ms--;
  } else if (millis() >= next40ms) {
      // test for 40ms time-out
      next40ms = millis() + 40; Task40ms = 3; // rset loop timer
  }

  //##############################################################################
  //
  // Handle serial communications
  //
  //##############################################################################
  if (Connected) {
    // Receive handler
    // If not connected to a remote device Rx_len == 0 and the following code will
    // not run. So nothing will come out of the serial port, other than connection
    // status messages, until a connection is established
    if (Rx_len > 0) {
      // data has been received from the remote device, so pass it to the serial port
      // reset the data in Rx_Buff[] to nulls as we go, then clear the flag
      if (Serial.availableForWrite() > 0) {
        // serial buffer has space in it so send a character
        Serial.write(Rx_Buff.ESPdata[Rx_Pnt]); Rx_Buff.ESPdata[Rx_Pnt] = 0;
        Rx_Pnt++; if (Rx_Pnt >= Rx_len) {Rx_len = 0;} // stop at end of data
      }
    }
  
    // Transmit handler
    // if not connected nothing will be broadcast over WiFi, and the link status will
    // be 'Disconnected', 'Listening'
    // A link has been established with a remote device so send data to it
    // At 115200 baud we could receive 14 bytes every millisecond from serial port
    // hence we will transmit the buffer contents every 2 ms, approx 30 bytes
    if ((Serial.available() > 0) && (Tx_len < 32)) {
      // serial data available so load up to 32 bytes into the WiFi Tx buffer
      if ((Tx_len == 0) || Tx_Hd) {
        // start with text header byte, or place one after Wii data
        Tx_Buff.ESPdata[Tx_len] = '$'; Tx_len++; Tx_Hd = false;
      }
      Tx_Char = Serial.read();  // get the char from the serial port
      Tx_Buff.ESPdata[Tx_len] = Tx_Char; Tx_len++;  // put it in the Tx buffer
      if (Tx_Echo) {Serial.write(Tx_Char);} // echo the received character on the serial port
    }
    if (millis() >= next2ms) {
      next2ms = millis() + 2; // reset the timer
      if (Tx_len > 0) {
        // the buffer contains data so add a checksum and send it
        t0 = micros();
        TxGetChecksum(Tx_len); Tx_Buff.ESPdata[Tx_len] = Tx_Chk; Tx_len++;
        esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, Tx_len); Tx_len = 0;
  //      Serial.println("WiFi Tx");
      }
      // do we need to send Nunchuk data in the next frame?
      if (Wii_Tx) {
        // yes we do, so it is placed at the front of the buffer, with Tx_len adjusted
        // and sent with the next frame in 2ms time
        PackI2CData(); Wii_Tx = false;
        Tx_Hd = true; // set flag to include '$' at head of text block
      }
    }
  }
//  else {
//    // In the 'Disconnected state we rely on the remote device initating the connection
//    if (Rx_len > 0) {
//      // decode the message from the remote device
//      decodeRxCmd();
//      if (WiFi_State == 1) {Tx_WhoID();}
//      if (WiFi_State == 2) {
//        // WiFi is re-connected, so send #Hinnn and clear flags
//        Tx_Hinnn();
//        Tx_len = 0; Wii_Tx = false; Tx = true; Tx_ErrCnt = 0;
//        Connected = true; Serial.println("WiFi Connected");
//      }
//      Rx_len = 0;
//    }
//  }
}

// --------------------------------------------------------------------------------

void OnDataRecv(uint8_t * mac, uint8_t *incomingData, uint8_t len) {
  // Callback when data is received
  // when not connected data will be accepted from any remote device
  // once connected the MAC address must match the registered device
  if (!Connected) {
    // currently not conencted so listen to any remote device that tries to connect
    // record device MAC address
    broadcastAddress[0] = mac[0]; broadcastAddress[1] = mac[1];
    broadcastAddress[2] = mac[2]; broadcastAddress[3] = mac[3];
    broadcastAddress[4] = mac[4]; broadcastAddress[5] = mac[5];
  } else {
    // connected to check that the data is from the established connection
    // ignore Rx data if MAC address does not match stored device MAC
    if (broadcastAddress[0] != mac[0]) {return;}
    if (broadcastAddress[1] != mac[1]) {return;}
    if (broadcastAddress[2] != mac[2]) {return;}
    if (broadcastAddress[3] != mac[3]) {return;}
    if (broadcastAddress[4] != mac[4]) {return;}
    if (broadcastAddress[5] != mac[5]) {return;}
  }
    
  while (Rx_len > 0) {
    // data has been received previously, so complete passing it to the serial port
    // before using the new data
    // reset the data in Rx_Buff[] to nulls as we go, then clear the flag
    if (Serial.availableForWrite() > 0) {
      Serial.write(Rx_Buff.ESPdata[Rx_Pnt]); Rx_Buff.ESPdata[Rx_Pnt] = 0;
      Rx_Pnt++; if (Rx_Pnt >= Rx_len) {Rx_len = 0;} // stop at end of data
    }
  }
  
  // we are now ready to handle the new packet of data
  memcpy(&Rx_Buff, incomingData, len);  // copies received data into Rx buffer
//  Serial.print("Bytes received: ");
//  Serial.println(len);
//  if (Connected) {
//    // whilst connected we may received '# Pings' so ignore them
//    if (len < 2) {
//      if (Rx_Buff.ESPdata[0] == '#') {Serial.print("#"); return;}
//      Serial.println("Rx Error - no Rx data"); return;
//    }
//  }
  
  if (Connected) {
    // WiFi connected so perform checksum check and if ok main loop will extract
    // data if Rx_len > 0
    Rx_len = len - 1; Rx_Pnt = 0;
    RxGetChecksum(Rx_len);  // generate a checksum from Rx data
    if (Rx_Chk != Rx_Buff.ESPdata[len - 1]) {
      Rx_len = 0; Serial.println("Rx Error - checksum");
    }
  } else {
    // if not connected this must be the first packet received, which is the name of
    // the remote device used to initiate a connection, so accept this name and 
    // enter connected transparent mode
    Rx_len = len;
    getRemoteName();
    Connected = true;
  }
}

// --------------------------------------------------------------------------------

void OnDataSent(uint8_t *mac_addr, uint8_t sendStatus) {
  // Callback when data is sent
  t1 = micros();  // record the loop back time to get here in microseconds
//  Serial.print("Send Status: ");
  if (sendStatus == 0){
    // No Tx errors detected so examine the loopback time
    // Tests on different packet data sizes revealed the following delays:
    // Data   Avg   Min    Max    Baud
    //   8    864   558   37228    74 kb/s avg
    //  16    896   573   34580   143 kb/s avg
    //  64    976   655   37116   656 kb/s avg
    //  96    1070  701   38033   897 kb/s avg
    // 128    1055  734   37447   1.2 Mb/s avg
    // 160    1039  786   39145   1.5 Mb/s avg
    // 192    1102  835   37813   1.7 Mb/s avg
    // 224    1232  870   37684   1.8 Mb/s avg
    // 250    1272  902   37289   1.9 Mb/s avg
    //
    // loop delay increases with packet size, as does efficiency
    // above baud rate is based on 10 bits/byte
//    t1_t0 = t1 - t0;
//    t1t0Cnt++; t1t0Sum += t1_t0;
//    if (t1_t0 < t1t0Min) {t1t0Min = t1_t0;}
//    if (t1_t0 > t1t0Max) {t1t0Max = t1_t0;}
//    Serial.println(String(DataTxLen) + "\t" + String(t1_t0) + "\t" + String(t1t0Sum/t1t0Cnt) + "\t" + String(t1t0Min) + "\t" + String(t1t0Max));
    Tx = true;
    Tx_CB = 1;      // set flag to indicate that a call back state has occured
    Tx_ErrCnt = 0;  // clear the error flag in case of a glitch
  }
  else{
    Serial.println("Tx Failed " + String(t1 - t0));
    Tx = false;
    Tx_CB = -1;  // set flag to indicate that a resend is required
    Tx_ErrCnt++; if (Tx_ErrCnt >= 5) {
      // too many send errors so break this connection
      Connected = false;
      // clear the MAC address of the previous remote device
      broadcastAddress[0] = 0; broadcastAddress[1] = 0;
      broadcastAddress[2] = 0; broadcastAddress[3] = 0;
      broadcastAddress[4] = 0; broadcastAddress[5] = 0;
      Serial.println("WiFi Disconnected");
    }
  }
}

// --------------------------------------------------------------------------------

void runPOST() {
  // called when the code first runs
  //##############################################################################
  //
  // Initialise the display
  //
  //##############################################################################
  Display_Init();
  Display_Intro();

  //##############################################################################
  //
  // Check battery health
  //
  //##############################################################################
  // first thing to do is to check the battery voltage
  Serial.println(F("Checking battery voltage..."));
  BatVol = analogRead(BatVolPin); // ignore the first reading
  delay(10); BatVol = analogRead(BatVolPin); // take the reading
  Serial.print(F("ADC = "));
  Serial.print(BatVol); Serial.print(F(" == "));
  // test for USB powered
  if (BatVol <= BatUSB) {
    BatPwr = false;
    Serial.print(float(BatVol)/BatCal); Serial.println(F(" volts"));
    Serial.println(F("USB port connection assumed")); BatQ = 5;
  } else {
    BatPwr = true;
    Serial.print((float(BatVol)/BatCal)); Serial.println(F(" volts"));
    if (BatVol < BatCritical) {
      // battery voltage is at or below the critical limit
      BatteryLow();
    } else {
      // rate the quality of the battery voltage 1 - 5
      // my fully charged battery was measured at 8.65v
      // so fully charged limit is: (((8.65 - 0.77)/2) * 1023/5) = 806
      BatQ = map(BatVol,BatCritical,806,1,5);
    }
  } 
  Serial.print(F("Battery quality = "));Serial.println(BatQ);

  //##############################################################################
  //
  // Initialise the Nunchuk controller
  //
  //##############################################################################
  // iniaitalise the Nunchuk and confirm its presence by setting a flag
  Serial.println(F("\nInitialising Wii controller over I2C..."));
  WiiInitialise(); Wii_Connected = true;
  if (WiiErrCnt > WiiErrRet) {
    Serial.println(F("Wii controller initialisation failed!"));
    Wii_Connected = false;
  } else {
    Serial.print(F("Wii controller initialised in ")); Serial.println(WiiErrCnt);
    ReadI2CIdent();   // determine the controller type
//    ReadI2CCalData(); // read calibration data
    XYOffAdj = false; WiiCal = 0; ReadI2CData();
  }

  // Init ESP-NOW
  if (esp_now_init() != 0) {
    Serial.println("Error initializing ESP-NOW");
  } else {Serial.println("\nESP-NOW Activated");}

  Serial.println(F("\nPOST completed."));
  Serial.println(F("Listening..."));
  delay(1000);
}

// --------------------------------------------------------------------------------

void Set_Defaults() {
  // define initial conditions for power-up and soft RESET
  BatAvg = 1023;              // preload averaging function with high value
  BatCnt = 0;                 // pointer to array values
  BatDel = 250;               // set a 10 second delay in 'life' prediction
  BatLast = 0;                // previous battery average reading
  BatPwr = false;             // true if battery supply detected, false if USB assumed
  BatSum = 20460;             // preload averaging sum with 20x max value
  BatT0 = 0;                  // time of first battery measurement in life prediction forecast
  BatTime$ = "--:--";         // time left displayed
  BatV0 = 0;                  // reference voltage used in prediction forecast
  BatV1 = 0;                  // lowest voltage used in prediction forecast
  BatVol = 0;                 // instantaneous battery voltage
  Connected = false;          // == true if a WiFi connection has been established
  DataTxLen = 250;            // length of data frame to send, min = 1, max = 250
  DataRxLen = 8;              // length of data frame received, min = 1, max = 250
  Dev_ID = -1;                // remote device ID, -1 default unknown
  DispCnt = 50;               // counter used in display updating
  DispClr = false;            // set == true to clear the memory contents
  DispDel = 0;                // display hold off counter, default = 0
  DispDisp = false;           // set = true to immediately display memory contents
  DispHold = false;           // set == true to stop display cycling
  DispNxt = 50;               // display switch timer, default = 50, 2 sec
  DispPnt = -1;               // points at the current display mode
  DispCX,DispCY;              // display cursor coordinates
  JoyXOff = 142;              // value to correct stick centre X offset
  JoyYOff = 137;              // value to correct stick centre Y offset
  next10ms = 0;               // 10ms loop timer trigger
  next40ms = 0;               // 40ms loop timer trigger
  nunchuk_address = 0x52;     // Nunchuk I2C address (0x52)
  Reboot = false;             // set true when reboot is in progress to stop all tasks
  reportEn = 0;               // reports data for testing if >= 1. Set this to 0 for transparent telemetry mode
  Rx_Chk = 0;                 // Rx checksum
  Rx_len = 0;                 // >0 when a packet has been received
  Rx_Pnt = 0;                 // receive pointer for data management
  STOP = false;               // set true when battery is critical
  sw1Cnt = 0;                 // button switch counter
  sw1LastState = HIGH;        // previous state of button switch, HIGH/LOW
  sw1State = HIGH;            // state of read button switch pin
  sw1Timer = 0;               // timer used to detemine button sequences
  t1t0Cnt = 0;                // number of t1-t0 readings
  t1t0Max = 0;                // max of t1-t0 times
  t1t0Min = 10000;            // min of t1-t0 times
  t1t0Sum = 0;                // sum of t1-t0 for averaging
  Task40ms = 0;               // task pointer for 40ms tasks
  Tx = false;                 // status of transmission, true == ok, false == failed
  Tx_CB = 1;                  // == 1 call back is clear to send, == -1 resend
  Tx_Char = 0;                // serial byte received from usb port
  Tx_Chk = 0;                 // Tx checksum
  Tx_Echo = false;            // ==true to echo serial data at port level, default false
  Tx_ErrCnt = 0;              // disconnects if error count is too high
  Tx_Hd = false;              // == true if a $ text header needs to be included in a block
  Tx_len = 0;                 // >0 when WiFi Tx buffer contains data to be sent
  WiFi_State = 0;             // WiFi connection state machine
  WiiCal = 0;                 // counter used in offset calibration
  Wii_Connected = false;      // true if a Wii contoller is detected
  Wii_Cnt = 0;                // counter used in Wii connection code
  WiiErrCnt = 0;              // number of attempts to initialse the Nunchuk
  WiiError = 1;               // used to detect transmission errors
  WiiTask = 0;                // initialisation task pointer
  WiiType = ' ';              // device type, 'N' == Nunchuk, 'C' == Classic
  Wii_Tx = false;             // ==true to load I2C data into the Tx buffer
  XYOffAdj = false;           // if == true adjust XY centre offsets

  ClearI2CData();
}

// --------------------------------------------------------------------------------
